
//needed imports. 
import scala.collection.mutable
import org.apache.spark.mllib.clustering.LDA
import org.apache.spark.mllib.linalg.{Vector, Vectors}
import org.apache.spark.rdd.RDD
import org.apache.spark.mllib.clustering.{DistributedLDAModel, LocalLDAModel}
import java.io._



// Load documents from text files, 1 document per file
//WholeTextFiles returns pairs (path, content) that's why we do the map and keep the second position. 
val corpus: RDD[String] = sc.wholeTextFiles("bbc/*/*.txt").map(_._2)


// Here we transform the text to lowercase, split the text by spaces, and filter the words of length < 3 and the numbers. 
val tokenized: RDD[Seq[String]] =
  corpus.map(_.toLowerCase.split("\\s")).map(_.filter(_.length > 3).filter(_.forall(java.lang.Character.isLetter)))

  
// Here we count how many times each word appears in the dataset by assigning a one to each word and then grouping by key (the word) adding the ones. Then we sort by the number of appearances. 
val termCounts: Array[(String, Long)] =
  tokenized.flatMap(_.map(_ -> 1L)).reduceByKey(_ + _).collect().sortBy(-_._2)
  
// We choose to remove the 20 more common words. 
val numStopwords = 20
// We also delete some commong words that don't give us any insight about the dataset. 
val badWords = Array("last", "first", "there","other","such", "many", "some","what", "while", "into", "when", "back", "those", "still")
//Remove the most commong words.
val vocabArray_temp: Array[String] =
  termCounts.takeRight(termCounts.size - numStopwords).map(_._1)
//Remove the words in the badWords List  
val vocabArray = vocabArray_temp.filter(word => !(badWords contains word))
  
//   vocab: Map term -> term index
// Create a hast table that has the word as key and the position of the word in the word-count vector as value.
val vocab: Map[String, Int] = vocabArray.zipWithIndex.toMap

// Convert documents into term count vectors
// Count the words of the tokenized RDD that are in the vocabulary in a hash table that has the index as key and the number of aparitions as value. 
// We then transform the hashtable to an sparse vector. 
val documents: RDD[(Long, Vector)] =
  tokenized.zipWithIndex.map { case (tokens, id) =>
    val counts = new mutable.HashMap[Int, Double]()
    tokens.foreach { term =>
      if (vocab.contains(term)) {
        val idx = vocab(term)
        counts(idx) = counts.getOrElse(idx, 0.0) + 1.0
      }
    }
    (id, Vectors.sparse(vocab.size, counts.toSeq))
  }

// We set the number of topics, the number of iterations and the optimizer to use. 
val numTopics = 5
val lda = new LDA().setK(numTopics).setMaxIterations(100)
lda.setOptimizer("em")

// We run the algorithm 
val ldaModel = lda.run(documents)

// Print topics, showing top-weighted 10 terms for each topic.
// We print the obtained topics and store them to a file called numberoftopic.txt.
var i = 0
val topicIndices = ldaModel.describeTopics(maxTermsPerTopic = 10)
topicIndices.foreach { case (terms, termWeights) =>
	i=i+1
	var pw = new PrintWriter(new File("BBCResults/topics/"+i+".txt" ))
  println("TOPIC: " + i)
  terms.zip(termWeights).foreach { case (term, weight) =>
    println(s"${vocabArray(term.toInt)}\t$weight")  
    pw.write(s"${vocabArray(term.toInt)}\t$weight")
	pw.write("\n")
  }
  pw.close
  println()
}




//When executing with em optimizer we obtain a DistributedLDAModel.  
val distLDA = ldaModel.asInstanceOf[DistributedLDAModel]

//The Local LDA Model is the one that has the method that gives us the topic for each document. 
val localLDA = distLDA.toLocal


//Load each of the documents we want to assign topics. And store one RDD with the id of each document. 
val corpus = sc.wholeTextFiles("bbc/*/*.txt")
val ids = corpus.zipWithIndex
val path_ID = ids.map(x => (x._1._1, x._2) )
val doc_ID = ids.map(x => (x._1._2,x._2))
val tokenized = doc_ID.map( {case (a,b) => (b,a.toLowerCase.split("\\s")) })

//This is the same as we did before to count the words of each document. 
val tokenized2 = tokenized.map({case (a,b)=> (a, b.filter(_.length > 3).filter(_.forall(java.lang.Character.isLetter)))})
val documents: RDD[(Long, Vector)] =
	  tokenized2.map { case (id, tokens) =>
		val counts = new mutable.HashMap[Int, Double]()
		tokens.foreach { term =>
		  if (vocab.contains(term)) {
			val idx = vocab(term)
			counts(idx) = counts.getOrElse(idx, 0.0) + 1.0
		  }
		}
		(id, Vectors.sparse(vocab.size, counts.toSeq))
	  }
	  
//get the topic distribution for each document.
val topics = localLDA.topicDistributions(documents)

//print the topics assigned to the first document of the RDD to be sure that we did it right. 
print(topics.first._1+" "+topics.first._2(0)+" "+topics.first._2(1)+" "+topics.first._2(2)+" "+topics.first._2(3)+" "+topics.first._2(4))


//store the assigned topics for each document in a file DocID.txt
topics.foreach({case(x) =>
	var pw = new PrintWriter(new File("BBCResults/documents/"+x._1+".txt" ))
	var str = x._1+" "+x._2(0)+" "+x._2(1)+" "+x._2(2)+" "+x._2(3)+" "+x._2(4)
	pw.write(str)
	pw.close	
	}
	)
	

//store the relation docID-Path for each document. 
path_ID.foreach({case(path,id)=>
	var pw = new PrintWriter(new File("BBCResults/ids/"+id+".txt" ))
	var str = path+" "+id
	pw.write(str)
	pw.close

})


